#!/usr/bin/env python
import sliderepl
sliderepl.SADeck.run()

### slide:: 1 Setup

from sqlalchemy import *

meta = MetaData(create_engine('sqlite://'))

users = Table('users', meta,
    Column('id', Integer, primary_key=True),
    Column('name', String(50))
)

addresses = Table('addresses', meta,
    Column('id', Integer, primary_key=True),
    Column('email', String(50)),
    Column('user_id', Integer, ForeignKey('users.id'))
)

meta.create_all()

ins = users.insert()

ins.execute([
   {'name':'jack'},
   {'name':'ed'},
   {'name':'wendy'},
   {'name':'mary'}
])

ins = addresses.insert()
ins.execute([
   {'user_id':1, 'email':'jack@jack.com'},
   {'user_id':2, 'email':'ed@yahoo.com'},
   {'user_id':2, 'email':'ed@msn.com'},
   {'user_id':3, 'email':'wendy@nyt.com'},
])


from sqlalchemy.orm import *
Session = scoped_session(sessionmaker(transactional=True,autoflush=False))

class Base(object):
    def __init__(self, **kwargs):
        for key, value in kwargs.iteritems():
            setattr(self, key, value)
    def __repr__(self):
        cls = self.__class__
        return "%s(%s)" % (
            cls.__name__,
            ", ".join(["%s=%r" % (c, getattr(self, c)) for c in cls.c.keys()])
        )
    query = Session.query_property()

### slide:: 2

class User(Base):
    pass

class Address(Base):
    pass

mapper(User, users, properties={
   'addresses':relation(Address, backref='user', cascade='all, delete-orphan')
})

mapper(Address, addresses)

### slide:: 3

print User.query.all()

### slide:: 4
print User.query.filter(User.addresses.any(Address.email.like('%ed%'))).all()

### slide:: 5

print User.query.filter(~User.addresses.any(Address.email.like('%ed%'))).all()

### slide:: 6

print Address.query.filter(Address.user.has(User.name=='ed')).all()

### slide:: 7

print Address.query.filter(~Address.user.has(User.name=='ed')).all()

### slide:: 8

u1 = User.query.get(2)
a1 = Address.query.get(1)

### slide:: 9

print User.query.filter(User.addresses.contains(a1)).all()

### slide:: 10

print Address.query.filter(Address.user == u1).all()

### slide:: 11

print Address.query.with_parent(u1).all()

### slide:: 12

print User.query.join('addresses').filter(Address.email.like('%y%')).all()

### slide:: 13

print User.query.filter(User.addresses.any(Address.email.like('%y%'))).all()

### slide:: 14

print User.query.filter(  exists([Address.id]).where(Address.user_id==User.id).\
    where(Address.email.like('%y%'))).all()

### slide:: 15

print User.query.filter(User.id==select([Address.user_id]).\
        where(Address.email.like('%y%'))).all()

### slide:: 16

nodes = Table('nodes', meta,
   Column('id', Integer, primary_key=True),
   Column('data', String(50)),
   Column('parent_id', Integer, ForeignKey('nodes.id'))
)

nodes.create()

### slide:: 17

class Node(Base):
   pass

# the "one to many" remote_side is the default,
# so we only need remote_side on the "many to one" side
mapper(Node, nodes, properties={
    'children':relation(Node,
         backref=backref('parent', remote_side=nodes.c.id)
     )
     })

### slide:: 18

Session.clear()

n1 = Node(data='n1', children=[
    Node(data='n2'),
    Node(data='n3', children=[
        Node(data='n5'),
        Node(data='n6'),
        Node(data='n7'),
    ]),
    Node(data='n4', children=[
        Node(data='n8'),
        Node(data='n9'),
    ]),
    ])

Session.save(n1)
Session.flush()

### slide:: 19

# SQLA Query Expression

parents = nodes.alias('parents')
grandparents = nodes.alias('grandparents')

print nodes.select().\
    select_from(nodes.join(parents, nodes.c.parent_id==parents.c.id).join(grandparents, parents.c.parent_id==grandparents.c.id)).\
    where(and_(grandparents.c.data=='n1', parents.c.data=='n4', nodes.c.data=='n8')).\
    execute().fetchall()

### slide:: 20

# ORM Query (using join() to create aliases)

n8 = Node.query.filter(Node.data=='n8').\
    join('parent', aliased=True).filter(Node.data=='n4').\
    join('parent', aliased=True, from_joinpoint=True).filter(Node.data=='n1').one()

print n8

### slide:: 21

# 1. write an ORM query to select all leaf nodes 
# (hint: any() need not take any arguments)
print Node.query.filter(~Node.children.any()).all()


### slide:: 22

# 2. write an ORM query to select the children of both 'n3'
# and 'n4'.  'n3' and 'n4' can be matched by their name alone
print Node.query.filter(
    Node.parent.has(Node.data.in_(['n3','n4']))
  ).all()
  
### slide::

q = User.query.\
   add_column(select([func.count(Address.id)]).\
   where(User.id==Address.user_id).as_scalar())

for user, count in q:
    print user, count

### slide:: 24 ORM Aggregate Subquery

sel = select([Address.user_id, func.count(Address.id).label('count')]).group_by(Address.user_id).alias()
print User.query.select_from(users.outerjoin(sel, User.id==sel.c.user_id)).add_column(sel.c.count).all()

### slide::

print User.query.outerjoin(('addresses', sel)).add_column(sel.c.count).all()

### slide:: 25

# pure SQL.  disambiguate common column names using
# <tablename>_<columnname>
sql = """
    select
      users.id as users_id,
      users.name,
      addresses.id as addresses_id,
      addresses.user_id,
      addresses.email
    from users
    left outer join addresses on users.id=addresses.user_id
    order by users.name
"""

### slide:: 26

# load from SQL statement
for user, address in \
   User.query.from_statement(sql).add_entity(Address):
   print user, address

### slide:: 27

# load from result set
result = meta.bind.execute(sql)

for user, address in \
   User.query.add_entity(Address).instances(result):
    print user, address

print User.query.add_entity(Address).instances(
    meta.bind.execute(sql))

### slide:: 28

Session.clear()

print User.query.options(contains_eager('addresses')).\
    from_statement(sql)

### slide:: 29

sql = """SELECT
        users.id AS users_id, users.name AS users_name,
        addresses.id AS addresses_id, addresses.email AS addresses_email,
        addresses.user_id AS addresses_user_id, adrc.id AS adrc_id,
        adrc.email AS adrc_email, adrc.user_id AS adrc_user_id
    FROM
    users JOIN addresses ON users.id=addresses.user_id
    LEFT OUTER JOIN addresses AS adrc ON users.id=adrc.user_id
    WHERE addresses.email like '%y%'"""

Session.clear()

for user, address in User.query.options(contains_eager('addresses', alias='adrc')).add_entity(Address).from_statement(sql):
    print user, user.addresses, address

### slide:: end
